""" Provides access to core functionality of the OsiriX database.

There should only ever be one, but it is OK to create multiple pyOsiriX instances, they will
just contiain the same `osirixrpc_uid`.

Example usage:
    ```python
    import osirix

    database = osirix.current_browser()  # Get the current browser instance
    images_path = "/path/to/dicoms"  # Define the location of the images
    database.copy_files_into_database(images_path)  # Load the images (on separate process).
    ```
"""

from __future__ import annotations
from typing import Tuple, List

import numpy as np
from numpy.typing import NDArray

import osirixgrpc.browsercontroller_pb2 as browsercontroller_pb2

import osirix
from osirix.base import pyosirix_connection_check


class BrowserController(osirix.base.OsirixBase):
    def __repr__(self):
        return "BrowserController"

    @pyosirix_connection_check
    def copy_files_into_database(self, files: List[str]) -> None:
        """ Copy files into the Osirix database.

        Note that this will always copy files, rather than copy by link, therefore doubling the
        memory requirements. It is safe to delete the source Dicom data after import, if you wish
        to do so.

        Args:
            files (List[str]): The list of files to copy, as absolute paths.

        Example usage:
             ```python
             database = osirix.current_browser()
             images_path = "/path/to/dicoms"
             database.copy_files_into_database(images_path)
             ```
        """
        request = browsercontroller_pb2.BrowserControllerCopyFilesIfNeededRequest(
            browser=self.pb2_object, paths=files)
        response = self.osirix_service_stub.BrowserControllerCopyFilesIfNeeded(request)
        self.response_check(response)

    @pyosirix_connection_check
    def database_selection(self) -> Tuple[List[osirix.dicom.DicomStudy],
                                          List[osirix.dicom.DicomSeries]]:
        """ Queries the user selection of Dicom images.

        Returns:
            The selected Dicom study instances
            The selected Dicom series instances

        Example usage:
             ```python
             database = osirix.current_browser()
             studies, series = database.database_selection()
             ```
        """
        response = self.osirix_service_stub.BrowserControllerDatabaseSelection(self.pb2_object)
        self.response_check(response)

        selected_studies = []
        for study in response.studies:
            selected_studies.append(osirix.dicom.DicomStudy(self.osirix_service, study))

        selected_series = []
        for series in response.series:
            selected_series.append(osirix.dicom.DicomSeries(self.osirix_service, series))

        return selected_studies, selected_series

    @pyosirix_connection_check
    def open_viewer_2d(self, dicom_images: NDArray) -> osirix.viewer_controller.ViewerController:
        """ Open a 2D viewer for an array of osirix.dicom.DicomImage instances.

        Note that images will be loaded in the order in which they are presented as input here.
        Use the properties of DicomImage to ensure that they are loaded in order (e.g. sorted by
        the `slice_location` or `instance_number` properties).

        Args:
            dicom_images (NDArray): The array of images to display. Must be one dimensional and
                have at least one element.

        Returns:
            The resulting new viewer controller.

        Raises:
            ValueError: When the number of images in `dicom_images` is less than 1.
            ValueError: When `dicom_images` is multidimensional.
            ValueError: When instances other than DicomImage are provided in `dicom_images`.

        Example usage:
             ```python
             database = osirix.current_browser()
             studies, series = database.database_selection()
             if len(series) == 0:
                 raise ValueError("No dicom series selected.")
             database.open_viewer_2d(series[0].sorted_images())
             ```
        """
        dicom_images = np.array(dicom_images)
        if len(dicom_images) <= 0:
            raise ValueError("The number of input images must be one or greater.")
        if not dicom_images.ndim == 1:
            raise ValueError("The input array must be one dimensional.")
        for dicom_image in dicom_images:
            if not isinstance(dicom_image, osirix.dicom.DicomImage):
                raise ValueError("Only instances of DicomImage are permitted as input.")

        dicom_images = [dicom_image.pb2_object for dicom_image in dicom_images]
        frames = [browsercontroller_pb2.BrowserControllerOpenViewerFromImagesRequest.FrameImages(
            images=dicom_images)]
        request = browsercontroller_pb2.BrowserControllerOpenViewerFromImagesRequest(
            browser=self.pb2_object, frames=frames, movie=False)
        response = self.osirix_service_stub.BrowserControllerOpenViewerFromImages(request)
        self.response_check(response)
        return osirix.viewer_controller.ViewerController(self.osirix_service, response.viewer)

    @pyosirix_connection_check
    def open_viewer_4d(self, dicom_images: NDArray) -> osirix.viewer_controller.ViewerController:
        """ Open a 4D viewer for an array of arrays of osirix.dicom.DicomImage instances.

        As per the `open_viewer_2d` method, each frames will be presented in the order in which
        the DicomImage instances are presented to the function.

        Note that no pre-checks are performed to ensure OsiriX can actually display the data (e.g.
        same slice locations are provided for each frame).

        Args:
            dicom_images (NDArray): An array of shape (N_frames, N_images).

        Returns:
            The resulting new viewer controller.

        Raises:
            ValueError: When N_frames is less than 2.
            ValueError: When N_images is less than 1.
            ValueError: When `dicom_images` shape is not 2-dimensional.
            ValueError: When instances other than DicomImage are provided in the `dicom_images`.

        Example usage:
             ```python
             database = osirix.current_browser()
             studies, series = database.database_selection()
             if len(series) < 2:
                 raise ValueError("Need at least two selected series.")
             database.open_viewer_4d(np.array([s.sorted_images() for s in series]))
             ```
        """
        dicom_images = np.array(dicom_images)
        if not dicom_images.ndim == 2:
            raise ValueError("The input array must be two dimensional.")

        n_frames, n_images = dicom_images.shape
        if not n_frames > 1:
            raise ValueError("Number of input frames must be > 1. "
                             "For single frame use `open_viewer_2d`.")
        if not n_images > 0:
            raise ValueError("Number of input images must be 1 or greater.")

        frames = []
        for i in range(n_frames):
            frame = []
            for j in range(n_images):
                dicom_image = dicom_images[i, j]
                if not isinstance(dicom_image, osirix.dicom.DicomImage):
                    raise ValueError("Only instances of DicomImage are permitted as input.")
                frame.append(dicom_image.pb2_object)
            frames.append(
                browsercontroller_pb2.BrowserControllerOpenViewerFromImagesRequest.FrameImages(
                    images=frame))

        request = browsercontroller_pb2.BrowserControllerOpenViewerFromImagesRequest(
            browser=self.pb2_object, frames=frames, movie=True)
        response = self.osirix_service_stub.BrowserControllerOpenViewerFromImages(request)
        self.response_check(response)
        return osirix.viewer_controller.ViewerController(self.osirix_service, response.viewer)



""" Classes that represent the core objects within the OsiriX database.

All attributes for these classes are immutable.

Example usage:
    ```python
    import osirix
    from osirix.dicom import DicomStudy, DicomSeries

    database = osirix.current_browser()  # Get the current browser instance
    studies, series = database.database_selection()  # Get the current selection by the user
    for study in studies:
        print(f"This is a DicomStudy instance with name {obj.study_name}")
    for serie in series:
        print(f"This is a DicomSeries instance with description {obj.series_description}")
    ```
"""

from __future__ import annotations

import datetime
from typing import List

import osirix  # noqa
from osirix.base import pyosirix_connection_check  # noqa


class DicomStudy(osirix.base.OsirixBase):
    """ An instance representing a Dicom study within the OsiriX database.

    """
    def __repr__(self):
        return f"DicomStudy: " \
               f"{self.patient_id} " \
               f"({self.date.strftime('%d/%m/%Y')})"

    @property
    @pyosirix_connection_check
    def date(self) -> datetime.datetime:
        """ The date in which the Dicom data were acquired.
        """
        response = self.osirix_service_stub.DicomStudyDate(self.pb2_object)
        self.response_check(response)
        study_date = datetime.datetime(response.year,
                                       response.month,
                                       response.day,
                                       response.hour,
                                       response.minute,
                                       response.second)
        return study_date

    @property
    @pyosirix_connection_check
    def date_added(self) -> datetime.datetime:
        """ The date in which the Dicom data were added to the OsiriX database.
        """
        response = self.osirix_service_stub.DicomStudyDateAdded(self.pb2_object)
        self.response_check(response)
        date_added = datetime.datetime(response.year,
                                       response.month,
                                       response.day,
                                       response.hour,
                                       response.minute,
                                       response.second)
        return date_added

    @property
    @pyosirix_connection_check
    def date_of_birth(self) -> datetime.datetime:
        """ The patient date of birth for the Dicom study.
        """
        response = self.osirix_service_stub.DicomStudyDateOfBirth(self.pb2_object)
        self.response_check(response)
        date_of_birth = datetime.datetime(response.year,
                                          response.month,
                                          response.day)
        return date_of_birth

    @property
    @pyosirix_connection_check
    def institution_name(self) -> str:
        """ The name of the institution where the Dicom study was performed.
        """
        response = self.osirix_service_stub.DicomStudyInstitutionName(self.pb2_object)
        self.response_check(response)
        return response.institution_name

    @property
    @pyosirix_connection_check
    def patient_name(self) -> str:
        """ The patient name of the Dicom study
        """
        response = self.osirix_service_stub.DicomStudyName(self.pb2_object)
        self.response_check(response)
        return response.name

    @property
    @pyosirix_connection_check
    def number_of_images(self) -> int:
        """ The number of images within the Dicom study.
        """
        response = self.osirix_service_stub.DicomStudyNumberOfImages(self.pb2_object)
        self.response_check(response)
        return response.no_images

    @property
    @pyosirix_connection_check
    def patient_id(self) -> str:
        """ The patient ID of the Dicom study.
        """
        response = self.osirix_service_stub.DicomStudyPatientID(self.pb2_object)
        self.response_check(response)
        return response.patient_id

    @property
    @pyosirix_connection_check
    def patient_sex(self) -> str:
        """ The patient sex of the Dicom study.
        """
        response = self.osirix_service_stub.DicomStudyPatientSex(self.pb2_object)
        self.response_check(response)
        return response.patient_sex

    @property
    @pyosirix_connection_check
    def patient_uid(self) -> str:
        """ The patient UID of the Dicom study.
        """
        response = self.osirix_service_stub.DicomStudyPatientUID(self.pb2_object)
        self.response_check(response)
        return response.patient_uid

    @property
    @pyosirix_connection_check
    def performing_physician(self) -> str:
        """ The performing physician name for the Dicom study.
        """
        response = self.osirix_service_stub.DicomStudyPerformingPhysician(self.pb2_object)
        self.response_check(response)
        return response.performing_physician

    @property
    @pyosirix_connection_check
    def referring_physician(self) -> str:
        """ The referring physician name for the Dicom study.
        """
        response = self.osirix_service_stub.DicomStudyReferringPhysician(self.pb2_object)
        self.response_check(response)
        return response.referring_physician

    @property
    @pyosirix_connection_check
    def study_instance_uid(self) -> str:
        """ The study instance UID for the Dicom series.
        """
        response = self.osirix_service_stub.DicomStudyStudyInstanceUID(self.pb2_object)
        self.response_check(response)
        return response.study_instance_uid

    @property
    @pyosirix_connection_check
    def study_name(self) -> str:
        """ The study name for the Dicom series.
        """
        response = self.osirix_service_stub.DicomStudyStudyName(self.pb2_object)
        self.response_check(response)
        return response.study_name

    @property
    @pyosirix_connection_check
    def modalities(self) -> str:
        """ The modalities available within the Dicom study.

        Modalities are seperated by a double backslash (e.g. "PT\\CT")
        """
        response = self.osirix_service_stub.DicomStudyModalities(self.pb2_object)
        self.response_check(response)
        return response.modalities

    @property
    @pyosirix_connection_check
    def series(self) -> List[DicomSeries]:
        """ The Dicom Series instances associated with the study.

        This will ignore any series associated with the OsiriX ROI Dicom SRs.
        """
        response = self.osirix_service_stub.DicomStudySeries(self.pb2_object)
        self.response_check(response)
        series = []
        for serie in response.series:
            response = self.osirix_service_stub.DicomSeriesModality(serie)
            if response.modality == "SR":
                continue
            series.append(DicomSeries(self.osirix_service, serie))
        return series

    @property
    @pyosirix_connection_check
    def images(self) -> List[DicomImage]:
        """ The Dicom Image instances associated with the study.
        """
        response = self.osirix_service_stub.DicomStudyImages(self.pb2_object)
        self.response_check(response)
        images = []
        for image in response.images:
            images.append(DicomImage(self.osirix_service, image))
        return images

    @property
    @pyosirix_connection_check
    def paths(self) -> List[str]:
        """ The Dicom file absolute paths associated with the study.

        This will ignore any paths associated with the OsiriX ROI Dicom SRs.
        """
        images = self.images
        paths = [image.complete_path for image in images if image.modality != "SR"]
        return paths

    @property
    @pyosirix_connection_check
    def no_files(self) -> int:
        """ The number of files comprising the Dicom study.

        """
        response = self.osirix_service_stub.DicomStudyNoFiles(self.pb2_object)
        self.response_check(response)
        return response.no_files

    @property
    @pyosirix_connection_check
    def raw_no_files(self) -> int:
        """ The number of raw image files within the Dicom study.

        Note that this will include the OsiriX ROI SR files if any exist.  Using `no_files` provides
            the number of accessible images.
        """
        response = self.osirix_service_stub.DicomStudyRawNoFiles(self.pb2_object)
        self.response_check(response)
        return response.no_files


class DicomSeries(osirix.base.OsirixBase):
    """ An instance representing a Dicom series within the OsiriX database.

    """
    def __repr__(self):
        return f"DicomSeries: " \
               f"{self.series_description} " \
               f"(N = {len(self.paths)})"

    @property
    @pyosirix_connection_check
    def date(self) -> datetime.datetime:
        """ The date in which the series was acquired.
        """
        response = self.osirix_service_stub.DicomSeriesDate(self.pb2_object)
        self.response_check(response)
        series_date = datetime.datetime(response.year,
                                        response.month,
                                        response.day,
                                        response.hour,
                                        response.minute,
                                        response.second)
        return series_date

    @property
    @pyosirix_connection_check
    def images(self) -> List[DicomImage]:
        """ The Dicom Image instances associated with the series.
        """
        response = self.osirix_service_stub.DicomSeriesImages(self.pb2_object)
        self.response_check(response)
        images = []
        for image in response.images:
            images.append(DicomImage(self.osirix_service, image))
        return images

    @property
    @pyosirix_connection_check
    def modality(self) -> str:
        """ The modality of the Dicom series.
        """
        response = self.osirix_service_stub.DicomSeriesModality(self.pb2_object)
        self.response_check(response)
        return response.modality

    @property
    @pyosirix_connection_check
    def patient_name(self) -> str:
        """ The patient name of the Dicom series
        """
        response = self.osirix_service_stub.DicomSeriesName(self.pb2_object)
        self.response_check(response)
        return response.name

    @property
    @pyosirix_connection_check
    def number_of_images(self) -> int:
        """ The number of images within the Dicom series.
        """
        response = self.osirix_service_stub.DicomSeriesNumberOfImages(self.pb2_object)
        self.response_check(response)
        return response.number_of_images

    @property
    @pyosirix_connection_check
    def series_description(self) -> str:
        """ The series description.
        """
        response = self.osirix_service_stub.DicomSeriesSeriesDescription(self.pb2_object)
        try:
            self.response_check(response)
        except osirix.exceptions.GrpcException:
            return "unknown series description"
        return response.series_description

    @property
    @pyosirix_connection_check
    def series_instance_uid(self) -> int:
        """ The series instance UID.

        This seems to generate some unexpected results. It is better to access directly
            by reading the dicom files (via DicomSeries.paths) using the pydicom library.
        """
        response = self.osirix_service_stub.DicomSeriesSeriesInstanceUID(self.pb2_object)
        self.response_check(response)
        return response.series_instance_uid

    @property
    @pyosirix_connection_check
    def sop_class_uid(self) -> int:
        """ The series SOP class UID.
        """
        response = self.osirix_service_stub.DicomSeriesSeriesSOPClassUID(self.pb2_object)
        self.response_check(response)
        return response.series_sop_class_uid

    @property
    @pyosirix_connection_check
    def study(self) -> DicomStudy:
        """ The Dicom Study containing the series.
        """
        response = self.osirix_service_stub.DicomSeriesStudy(self.pb2_object)
        self.response_check(response)
        return DicomStudy(self.osirix_service, response.study)

    @property
    @pyosirix_connection_check
    def paths(self) -> List[str]:
        """ The Dicom file paths associated with the series.

        Returns:
            A list of absolute paths.
        """
        response = self.osirix_service_stub.DicomSeriesPaths(self.pb2_object)
        self.response_check(response)
        paths = [path for path in response.paths]
        return paths

    @property
    @pyosirix_connection_check
    def next_series(self) -> DicomSeries:
        """ The next Dicom Series in the OsiriX database.
        """
        response = self.osirix_service_stub.DicomSeriesNextSeries(self.pb2_object)
        self.response_check(response)
        return DicomSeries(self.osirix_service, response.series)

    @property
    @pyosirix_connection_check
    def previous_series(self) -> DicomSeries:
        """ The previous Dicom Series in the OsiriX database.
        """
        response = self.osirix_service_stub.DicomSeriesPreviousSeries(self.pb2_object)
        self.response_check(response)
        return DicomSeries(self.osirix_service, response.series)

    @pyosirix_connection_check
    def sorted_images(self) -> List[DicomImage]:
        """ Dicom image instances sorted by OsiriX user preferences.

        These preferences are determined in OsiriX Preferences / Database / Sorting.

        Returns:
           A list of sorted DicomImage instances.
        """
        response = self.osirix_service_stub.DicomSeriesSortedImages(self.pb2_object)
        self.response_check(response)
        sorted_images = []
        for sorted_image in response.sorted_images:
            sorted_images.append(DicomImage(self.osirix_service, sorted_image))
        return sorted_images


class DicomImage(osirix.base.OsirixBase):
    """ An instance representing a Dicom image within the OsiriX database.

    """
    def __repr__(self):
        return f"DicomImage: " \
               f"({self.instance_number} / {len(self.series.paths)})"

    @property
    @pyosirix_connection_check
    def date(self) -> datetime.datetime:
        """ The date in which the Dicom image were acquired.
        """
        response = self.osirix_service_stub.DicomImageDate(self.pb2_object)
        self.response_check(response)
        image_date = datetime.datetime(response.year,
                                       response.month,
                                       response.day,
                                       response.hour,
                                       response.minute,
                                       response.second)
        return image_date

    @property
    @pyosirix_connection_check
    def instance_number(self) -> int:
        """ The instance number for the Dicom image.
        """
        response = self.osirix_service_stub.DicomImageInstanceNumber(self.pb2_object)
        self.response_check(response)
        return response.instance_number

    @property
    @pyosirix_connection_check
    def modality(self) -> str:
        """ The modality of the Dicom image.
        """
        response = self.osirix_service_stub.DicomImageModality(self.pb2_object)
        self.response_check(response)
        return response.modality

    @property
    @pyosirix_connection_check
    def number_of_frames(self) -> int:
        """ The number of frames within the Dicom image.
        """
        response = self.osirix_service_stub.DicomImageNumberOfFrames(self.pb2_object)
        self.response_check(response)
        return response.number_of_frames

    @property
    @pyosirix_connection_check
    def series(self) -> DicomSeries:
        """ The DicomSeries instance in which the image is contained.
        """
        response = self.osirix_service_stub.DicomImageSeries(self.pb2_object)
        self.response_check(response)
        return DicomSeries(self.osirix_service, response.series)

    @property
    @pyosirix_connection_check
    def slice_location(self) -> float:
        """ The slice location of the image.

        Note that this is never perfect; it can be better to read the source dicom instance and use
            the ImagePositionPatient parameter.
        """
        response = self.osirix_service_stub.DicomImageSliceLocation(self.pb2_object)
        self.response_check(response)
        return response.slice_location

    @property
    @pyosirix_connection_check
    def complete_path(self) -> str:
        """ The path of the source Dicom file.
        """
        response = self.osirix_service_stub.DicomImageCompletePath(self.pb2_object)
        self.response_check(response)
        return response.complete_path

    @property
    @pyosirix_connection_check
    def height(self) -> int:
        """ The number of rows in the image
        """
        response = self.osirix_service_stub.DicomImageHeight(self.pb2_object)
        self.response_check(response)
        return response.height

    @property
    @pyosirix_connection_check
    def width(self) -> int:
        """ The number of columns in the image
        """
        response = self.osirix_service_stub.DicomImageWidth(self.pb2_object)
        self.response_check(response)
        return response.width

    @property
    @pyosirix_connection_check
    def sop_instance_uid(self) -> str:
        """ The SOP instance UID for the image.
        """
        response = self.osirix_service_stub.DicomImageSOPInstanceUID(self.pb2_object)
        self.response_check(response)
        return response.sop_instance_uid



""" Provides functionality for the images displayed in a 2D OsiriX viewer.

Example usage:
     ```python
     import osirix
     import matplotlib.pyplot as plt

     viewer = osirix.frontmost_viewer()  # Raises GrpcException error if no viewer is available.
     pix = viewer.cur_dcm()  # Get the currently displayed DCMPix object
     plt.imshow(pix.image, cmap = "gray")  # Display the image data (pix.image is a 2D numpy array)
     plt.show()
     ```
"""

from __future__ import annotations
from typing import Dict, Tuple

from numpy.typing import NDArray
import numpy as np

import osirixgrpc.dcmpix_pb2 as dcmpix_pb2

import osirix
from osirix.base import pyosirix_connection_check


class DCMPix(osirix.base.OsirixBase):
    def __repr__(self):
        return f"DCMPix: " \
               f"{self.slice_location:.2f} " \
               f"{self.shape}"

    @property
    @pyosirix_connection_check
    def is_rgb(self) -> bool:
        """ Is the image data red-green-blue? If `False` must be greyscale.
        """
        response = self.osirix_service_stub.DCMPixIsRGB(self.pb2_object)
        self.response_check(response)
        return response.is_rgb

    @property
    @pyosirix_connection_check
    def slice_location(self) -> float:
        """ The slice location of the image.

        NOTE: It can be more accurate to get the slice location by loading the dicom file via the
        `source_file` property (using pydicom for example), and using the ImagePositionPatient tag.
        """
        response = self.osirix_service_stub.DCMPixSliceLocation(self.pb2_object)
        self.response_check(response)
        return float(response.slice_location)

    @property
    @pyosirix_connection_check
    def orientation(self) -> NDArray:
        """ The orientation of the image.

        NOTE: It can be more accurate to get the slice location by loading the dicom file via the
        `source_file` property (using pydicom for example), and using the ImageOrientationPatient
        tag.
        """
        response = self.osirix_service_stub.DCMPixOrientation(self.pb2_object)
        self.response_check(response)
        orientation = []
        for o in response.orientation:
            orientation.append(float(o))
        return np.array(orientation)

    @property
    @pyosirix_connection_check
    def origin(self) -> Tuple[float, float, float]:
        """ The origin of the image (x, y, z).

        NOTE: It can be more accurate to get the slice location by loading the dicom file via the
        `source_file` property (using pydicom for example), and using the ImagePositionPatient tag.
        """
        response = self.osirix_service_stub.DCMPixOrigin(self.pb2_object)
        self.response_check(response)
        return (float(response.origin_x),
                float(response.origin_y),
                float(response.origin_z))

    @property
    @pyosirix_connection_check
    def pixel_spacing(self) -> Tuple[float, float]:
        """ The pixel spacing of the image (order: rows, columns)
        """
        response = self.osirix_service_stub.DCMPixSpacing(self.pb2_object)
        self.response_check(response)
        return float(response.spacing_rows), float(response.spacing_columns)

    @property
    @pyosirix_connection_check
    def shape(self) -> Tuple[int, int]:
        """ The pixel shape of the image (order: rows, columns)
        """
        response = self.osirix_service_stub.DCMPixShape(self.pb2_object)
        self.response_check(response)
        return int(response.rows), int(response.columns)

    @property
    @pyosirix_connection_check
    def source_file(self) -> str:
        """ The source file of the image on the host machine.
        """
        response = self.osirix_service_stub.DCMPixSourceFile(self.pb2_object)
        self.response_check(response)
        return response.source_file

    @property
    @pyosirix_connection_check
    def image(self) -> NDArray:
        """ The image data as a Numpy array.

        If the image is RGB format, then the shape will be (rows, columns, 4), whereas if the image
        is greyscale format it will be shape (rows, columns) (see `is_rgb` property).
        """
        response = self.osirix_service_stub.DCMPixImage(self.pb2_object)
        self.response_check(response)
        if response.is_argb:
            image_array = np.array(response.image_data_argb).reshape((response.rows,
                                                                      response.columns,
                                                                      4))
        else:
            image_array = np.array(response.image_data_float).reshape((response.rows,
                                                                       response.columns))

        return image_array

    @image.setter
    @pyosirix_connection_check
    def image(self, image_arr: NDArray):
        """ The image data as a Numpy array.

        If the image is RGB format, then the shape will be (rows, columns, 4), whereas if the image
        is greyscale format it will be shape (rows, columns) (see `is_rgb` property).
        """
        is_rgb = self.is_rgb
        if is_rgb:
            request = dcmpix_pb2.DCMPixSetImageRequest(pix=self.pb2_object,
                                                       image_data_argb=image_arr.ravel().tolist())

        else:
            request = dcmpix_pb2.DCMPixSetImageRequest(pix=self.pb2_object,
                                                       image_data_float=image_arr.ravel().tolist())

        response = self.osirix_service_stub.DCMPixSetImage(request)
        self.response_check(response)

    @pyosirix_connection_check
    def compute_roi(self, roi: osirix.roi.ROI) -> Dict:
        """ Compute some statistics from an ROI contained within the image.

        Note that these are calculated internally by OsiriX.  For more refinement on how statistics
        are calculated we suggest that it is better to create your own tools (based on the
        `scipy.stats` library for example).

        Args:
            roi (osirix.roi.ROI): The region of interest from which to compute the statistics.

        Returns:
            A dictionary containing the following key-value pairs: "mean", "std", "min", "max",
                "skewness", "kurtosis"
        """
        request = dcmpix_pb2.DCMPixComputeROIRequest(pix=self.pb2_object, roi=roi.pb2_object)
        response = self.osirix_service_stub.DCMPixComputeROI(request)
        self.response_check(response)
        roi_dict = {
            "mean": response.mean,
            "total": response.total,
            "std_dev": response.std_dev,
            "min": response.min,
            "max": response.max,
            "skewness": response.skewness,
            "kurtosis": response.kurtosis
        }
        return roi_dict

    @pyosirix_connection_check
    def convert_to_bw(self, mode: int = 3) -> None:
        """ Convert the image to greyscale.

        Args:
            mode (int): 0 = use red channel, 1 = use green channel, 2 = use blue channel, and
                3 = merge.
        """
        if mode not in [0, 1, 2, 3]:
            raise ValueError("Mode must 0, 1, 2, or 3")
        request = dcmpix_pb2.DCMPixConvertToBWRequest(pix=self.pb2_object,
                                                      bw_channel=mode)
        response = self.osirix_service_stub.DCMPixConvertToBW(request)
        self.response_check(response)

    @pyosirix_connection_check
    def convert_to_rgb(self, mode: int = 3) -> None:
        """ Convert the image to RGB.

        Args:
            mode (int): 0 = create red channel, 1 = create green channel, 2 = create blue channel,
                and 3 = create all channels.
        """
        request = dcmpix_pb2.DCMPixConvertToRGBRequest(pix=self.pb2_object, rgb_channel=mode)
        response = self.osirix_service_stub.DCMPixConvertToRGB(request)
        self.response_check(response)

    @pyosirix_connection_check
    def get_map_from_roi(self, roi: osirix.roi.ROI) -> NDArray:
        """ Create a mask from an input ROI based on the image.

        Args:
            roi (osirix.roi.ROI): The ROI from which to extract the mask.

        Returns:
            The mask as a 2-dimensional binary array with shape (rows, columns).
        """
        request = dcmpix_pb2.DCMPixGetMapFromROIRequest(pix=self.pb2_object,
                                                        roi=roi.pb2_object)
        response = self.osirix_service_stub.DCMPixGetMapFromROI(request)
        self.response_check(response)
        mask = np.array(response.map).reshape(response.rows, response.columns)
        return mask

    @pyosirix_connection_check
    def get_roi_values(self, roi: osirix.roi.ROI) -> Tuple[NDArray, NDArray, NDArray]:
        """ Extract the pixel values within a region of interest.

        Args:
            roi (osirix.roi.ROI): The ROI from which to extract values.

        Returns:
            The row indices of the extracted values.
            The columns indices of the extracted values.
            The extracted pixel values.
        """
        request = dcmpix_pb2.DCMPixROIValuesRequest(pix=self.pb2_object, roi=roi.pb2_object)
        response = self.osirix_service_stub.DCMPixROIValues(request)
        self.response_check(response)
        rows = np.array(response.row_indices)
        columns = np.array(response.column_indices)
        values = np.array(response.values)
        return rows, columns, values

    @pyosirix_connection_check
    def image_obj(self) -> osirix.dicom.DicomImage:
        """ The `DicomImage` instance from which the image was derived.

        Returns:
            The image instance.
        """
        response = self.osirix_service_stub.DCMPixDicomImage(self.pb2_object)
        self.response_check(response)
        return osirix.dicom.DicomImage(self.osirix_service, response.dicom_image)

    @pyosirix_connection_check
    def series_obj(self) -> osirix.dicom.DicomSeries:
        """ The `DicomSeries` instance from which the image was derived.

        Returns:
            The series instance.
        """
        response = self.osirix_service_stub.DCMPixDicomSeries(self.pb2_object)
        self.response_check(response)
        return osirix.dicom.DicomSeries(self.osirix_service, response.dicom_series)

    @pyosirix_connection_check
    def study_obj(self) -> osirix.dicom.DicomStudy:
        """ The `DicomStudy` instance from which the image was derived.

        Returns:
            The study instance.
        """
        response = self.osirix_service_stub.DCMPixDicomStudy(self.pb2_object)
        self.response_check(response)
        return osirix.dicom.DicomStudy(self.osirix_service, response.dicom_study)



""" The main osirix module.

This module contains some core methods for access to the core OsiriX objects.

Typical usage example:

    import osirix
    frontmost_viewer = osirix.frontmost_viewer()
    vr_controller = osirix.vr_controller()
"""

from __future__ import annotations

__version__ = "0.2.1-beta5"

__author__ = "Timothy Sum Hon Mun & Matthew D Blackledge"

import os
import json
import warnings
from typing import List, AnyStr, Tuple, Dict

from . import base
from . import browser_controller
from . import dcm_pix
from . import dicom
from . import exceptions
from . import osirix_utilities
from . import roi
from . import viewer_controller
from . import vr_controller

# An instance of the automatically established OsiriX class.
# Do not change or access directly.
_osirix = None


def plugin_support_directory() -> AnyStr:
    """ Determine the location of the OsiriXgrpc plugin support directory.

    Returns:
        The plugin support directory as an absolute path.
    """
    home = os.path.expanduser("~")
    support_directory = os.path.join(home,
                                     "Library",
                                     "Application Support",
                                     "OsirixGRPC")
    return support_directory


def allowed_plugin_types() -> Dict:
    """ The plugin types allowed, both integer and string names.

    Returns:
        Dict: The allowed plugin types.
    """
    return {1: "gRPCImageRenderTool",
            2: "gRPCROITool",
            3: "gRPCVolumeRenderTool",
            4: "gRPCDatabaseTool"}


def osirixgrpc_port_preference() -> int | None:
    """ Extract the preferred port set by the user in the OsiriXgrpc plugin.

    It will iterate through the ports set-up by the user in order, and return
    the first open one. If you want more refinement (e.g. using a different
    port), then you must set up your own service.

    Returns:
        The preferred port. If one is not found or none are open, the `None`
            is returned.
    """
    port = None
    support_directory = plugin_support_directory()
    server_configs = os.path.join(support_directory,
                                  "server_configs.json")
    if not os.path.exists(server_configs):
        return None

    with open(server_configs, "rb") as fhandle:
        server_configs_dict = json.load(fhandle)
        for item in server_configs_dict:
            if item["active"] == "YES":
                port = item["port"]
    return port


def install_script(name: str, plugin_type: int, path: str, interpreter_path: str = None,
                   blocking: bool = False) -> None:
    """ Install a plugin into OsiriX.

    Args:
        name (str): The name of the plugin.
        plugin_type (int): The plugin type: 1 = Image, 2 = ROI, 3 = VR, 4 = Database.
        path (str): The path of the script that will be run.
        interpreter_path (str): The interpreter path to run the script.
        blocking (bool): Whether to block OsiriX while running the script.
    """
    # Perform checks.
    if plugin_type not in allowed_plugin_types().keys():
        raise ValueError("Plugin type must be 1 (Image), 2 (ROI), 3 (VR), or 4 (Database).")

    # Open the script configs.
    support_directory = plugin_support_directory()
    script_configs_path = os.path.join(support_directory,
                                       "script_configs.json")
    if not os.path.exists(script_configs_path):
        raise FileNotFoundError("No script configs file found.")

    with open(script_configs_path, "r") as fhandle:
        script_configs = json.load(fhandle)

    # Check that the name is valid.
    for script_config in script_configs:
        if script_config["name"] == name and script_config["type"] == plugin_type:
            raise ValueError("Duplicate script found")

    # Append the new script information
    blocking_str = "NO"
    if blocking:
        blocking_str = "YES"
    script_config_new = {"name": name,
                         "type": allowed_plugin_types()[plugin_type],
                         "path": path,
                         "interpreterPath": interpreter_path,
                         "blocking": blocking}

    # Save to file
    with open(script_configs_path, "w") as fhandle:
        json.dump(script_config_new, fhandle)


def global_osirix_instance() -> osirix_utilities.Osirix:
    """ Get the global osirix instance.

    This should be set-up on first import of the osirix package, and uses the
    default set by the OsiriXgrpc plugin.

    Returns:
        The global osirix instance used to access the core OsiriX functions.
    """
    global _osirix

    if _osirix is None:
        port = osirixgrpc_port_preference()
        if port is not None:
            service = osirix_utilities.OsirixService(domain="127.0.0.1",
                                                     port=port,
                                                     max_send_message_length=512 * 1024 * 1024,
                                                     max_receive_message_length=512 * 1024 * 1024)
            try:
                service.start_service()
            except exceptions.GrpcException:
                print("Could not start global service."
                      "OsiriX may not be running and you may need to start it.")
            _osirix = osirix_utilities.Osirix(service)
    return _osirix


def _establish_global_osirix_instance():
    """ Establish that a connection can be established on import.
    """
    if global_osirix_instance() is None:
        warnings.warn("No connection with OsiriX established on initialization."
                      "You may need to set one up yourself.")


# The following functions are defined for ease of access.
def current_browser() -> browser_controller.BrowserController:
    """ The main Dicom browser instance.

    Returns:
        The main Dicom browser.

    Example:
        ```python
        import osirix
        browser = osirix.current_browser()
        ```
    """
    osirix_ = global_osirix_instance()
    if osirix_ is None:
        raise exceptions.GrpcException("No connection could be established with OsiriX")
    return osirix_.current_browser()


def frontmost_viewer() -> viewer_controller.ViewerController:
    """ The front-most 2D viewer.

    If there are more than one, look for the viewer highlighted by a red boundary.

    Returns:
        The viewer controller instance.

    Example:
        ```python
        import osirix
        viewer = osirix.frontmost_viewer()
        ```
    """
    osirix_ = global_osirix_instance()
    if osirix_ is None:
        raise exceptions.GrpcException("No connection could be established with OsiriX")
    return osirix_.frontmost_viewer()


def displayed_2d_viewers() -> List[viewer_controller.ViewerController]:
    """ All currently active 2D viewers.

    Returns:
        A list of all displayed 2D viewers.

    Example:
        ```python
        import osirix
        viewers = osirix.displayed_2d_viewers()
        ```
    """
    osirix_ = global_osirix_instance()
    if osirix_ is None:
        raise exceptions.GrpcException("No connection could be established with OsiriX")
    return osirix_.displayed_2d_viewers()


def frontmost_vr_controller() -> vr_controller.VRController:
    """ The front-most 3D viewer.

    Returns:
        The VR controller instance.

    Example:
        ```python
        import osirix
        vr_controller = osirix.frontmost_vr_controller()
        ```
    """
    osirix_ = global_osirix_instance()
    if osirix_ is None:
        raise exceptions.GrpcException("No connection could be established with OsiriX")
    return osirix_.frontmost_vr_controller()


def displayed_vr_controllers() -> List[vr_controller.VRController]:
    """ All currently active 3D viewers.

    Returns:
        A list of all displayed 3D viewers.

    Example:
        ```python
        import osirix
        vr_controllers = osirix.displayed_vr_controllers()
        ```
    """
    osirix_ = global_osirix_instance()
    if osirix_ is None:
        raise exceptions.GrpcException("No connection could be established with OsiriX")
    return osirix_.displayed_vr_controllers()


def osirix_version() -> Tuple[str, str]:
    """ The version of OsiriX being linked to.

    This function can also be used as a 'ping' to check that a connection is established.

    Returns:
        str: The OsiriX version.

    Example:
        ```python
        import osirix
        version = osirix.osirix_version()
        ```
    """
    osirix_ = global_osirix_instance()
    if osirix_ is None:
        raise exceptions.GrpcException("No connection could be established with OsiriX")
    return osirix_.osirix_version()


_establish_global_osirix_instance()



""" The core functionality of pyOsiriX (base class and service class).

"""

from __future__ import annotations

from typing import Union

import osirixgrpc
from osirixgrpc import types_pb2

import osirix
from osirix.exceptions import GrpcException

# Define the OsiriXgrpc types allowed.
Pb2Type = Union[types_pb2.DicomStudy, types_pb2.DicomSeries, types_pb2.DicomImage, types_pb2.ROI,
                types_pb2.ROIVolume, types_pb2.DCMPix, types_pb2.BrowserController,
                types_pb2.ViewerController, types_pb2.VRController, None]


def pyosirix_connection_check(func):
    """ A decorator function that checks an OsiriX connection is established.  """
    def wrapper(self, *args, **kwargs):
        if self.osirix_service.check_connection():
            return func(self, *args, **kwargs)
        else:
            raise GrpcException("No connection with OsiriX available.")
    return wrapper


class OsirixBase:
    """ A base class to be used for all pyOsiriX classes. Only subclasses must be created.

    """
    def __init__(self, osirix_service: osirix.osirix_utilities.OsirixService,
                 pb2_object: Pb2Type = None):
        self._osirix_service = osirix_service
        self._osirix_service_stub = osirix_service.osirix_service_stub
        self.pb2_object = pb2_object  # The protobuf object created using osirixgrpc (private)

    def __eq__(self, other):
        """ We let osirixgrpc handle this.
        """
        return self.pb2_object == other.pb2_object

    @property
    def osirix_service(self) -> osirix.osirix_utilities.OsirixService:
        """ The OsiriX service from which the instance was created.
        """
        return self._osirix_service

    @property
    def osirix_service_stub(self) -> osirixgrpc.osirix_pb2_grpc.OsiriXServiceStub:
        """ The service stub contained with service. Provided for convenient/fast access.
        """
        return self._osirix_service_stub

    @staticmethod
    def response_check(grpc_response) -> None:
        """ Checks whether the status of a response and raises an appropriate error.

        Args:
            grpc_response: Response returned by Osirix service. All returned OsiriXgrpc messages
                contain a `status` (1 for success, 0 otherwise).  Where the status is 0, an
                additional `message` is provided, outlining what went wrong.
        """
        if grpc_response.status.status == 1:
            pass
        elif grpc_response.status.message != "":
            raise osirix.exceptions.GrpcException(grpc_response.status.message)
        else:
            raise osirix.exceptions.GrpcException("Bad response. Please contact developer.")



""" Functionality for the regions of interest in 2D (`ROI`) and 3D (`ROIVolume`) viewers.

The following `ROI` types are available in pyOsiriX, seperated into 2 core groups:

1. Brush ROIs: Described by a 2D array of boolean values (0 = outside ROI, 1 = within ROI).

    - tPlain: The only brush ROI type.

2. Point-based ROIs: Described by a set of N 2-element vertices as a shape (N, 2) array.

    - tMeasure: A ruler measurement. N = 2.
    - tROI: A square ROI. N = 4. Created by rectangle.
    - tOval: An oval-based ROI. No limit on N. Created by rectangle.
    - tOPolygon: An open polygon. No limit on N.
    - tCPolygon: A closed polygon. No limit on N.
    - tAngle: An angle measurement. N = 3.
    - tText: A text box. Created by rectangle.
    - tArrow: An arrow. N = 2 (arrow head then tail).
    - tPencil: A closed polygon. No limit it N.
    - t2DPoint: A single point ROI. Created by rectangle.
    - tAxis: An axis-based ROI for bi-dimensional measurement. N = 4.
    - tDynAngle: The angle between two lines. N = 4.
    - tTAGT: A pair of oriented perpendicular lines for measuring distance. N = 6.
    - tBall: A circular ROI representing a slice of a 3D ball ROI. No limit on N.
        It is not possible to create a new tBall ROI with pyOsiriX.
    - tOvalAngle: An oval with an additional angle specifier. No limit on N. Created by
        rectangle.

Example usage:
    ```python
    import osirix
    import matplotlib.pyplot as plt

    viewer = osirix.frontmost_viewer()  # Raises GrpcException error if no viewer is available.
    rois = viewer.selected_rois()  # Get the currently selected ROI object.
    if len(rois) == 0:
        raise ValueError("No ROIs found.")
    roi = rois[0]
    roi_type = roi.itypes()[roi.itype]  # Get the selected ROI type.
    if roi_type == "tPlain":
        raise ValueError("Cannot plot mask (brush) ROI vertices")
    plt.plot(roi.points[:, 0], roi.points[:, 1], "ko")  # Plot the ROI vertices.
    plt.show()
    ```
"""

from __future__ import annotations

import warnings
from typing import Tuple, Dict

import numpy as np
from numpy.typing import NDArray

import osirixgrpc.roi_pb2 as roi_pb2
import osirixgrpc.roivolume_pb2 as roivolume_pb2

import osirix
from osirix.base import pyosirix_connection_check


class ROIVolume(osirix.base.OsirixBase):
    """ Represents a single volumetric region of interest in a 3D viewer (`VRController`)

    """
    def __init__(self, vr_controller: osirix.vr_controller.VRController,
                 osirix_service: osirix.osirix_utilities.OsirixService, *args, **kwargs):
        super(ROIVolume, self).__init__(osirix_service, *args, **kwargs)
        self._vr_controller = vr_controller

    def __repr__(self):
        return f"ROIVolume: " \
               f"{self.name}"

    @property
    @pyosirix_connection_check
    def vr_controller(self) -> osirix.vr_controller.VRController:
        """ The 3D volume render window in which this ROI is shown.
        """
        return self._vr_controller

    @property
    @pyosirix_connection_check
    def texture(self) -> bool:
        """ Whether the volume ROI is textured
        """
        response = self.osirix_service_stub.ROIVolumeTexture(self.pb2_object)
        self.response_check(response)
        return response.texture

    @texture.setter
    @pyosirix_connection_check
    def texture(self, texture: bool):
        """ Whether the volume ROI is textured
        """
        request = roivolume_pb2.ROIVolumeSetTextureRequest(roi=self.pb2_object,
                                                           texture=texture)
        response = self.osirix_service_stub.ROIVolumeSetTexture(request)
        self.response_check(response)

    @property
    @pyosirix_connection_check
    def visible(self) -> bool:
        """ Whether the volume ROI is visible in the 3D viewer
        """
        response = self.osirix_service_stub.ROIVolumeVisible(self.pb2_object)
        self.response_check(response)
        return response.visible

    @visible.setter
    @pyosirix_connection_check
    def visible(self, set_visible: bool):
        """ Whether the volume ROI is visible in the 3D viewer
        """
        if self.visible and not set_visible:
            self.vr_controller.hide_roi_volume(self)
        if not self.visible and set_visible:
            self.vr_controller.display_roi_volume(self)

    @property
    @pyosirix_connection_check
    def name(self) -> str:
        """ The name of the ROI volume.
        """
        response = self.osirix_service_stub.ROIVolumeName(self.pb2_object)
        self.response_check(response)
        return response.name

    @property
    @pyosirix_connection_check
    def color(self) -> Tuple[float, float, float]:
        """ The color of the volume ROI as a (r, g, b) tuple (each channel in range 0-1)
        """
        response = self.osirix_service_stub.ROIVolumeColor(self.pb2_object)
        self.response_check(response)
        return response.r, response.g, response.b

    @color.setter
    @pyosirix_connection_check
    def color(self, color: Tuple[float, float, float]):
        """ The color of the volume ROI as a (r, g, b) tuple (each channel in range 0-1)
        """
        r, g, b = color
        if r < 0 or r > 1:
            r = np.clip(r, 0, 1)
            warnings.warn(f"Red color outside range 0 - 1. Clipping to {r}")
        if g < 0 or g > 1:
            g = np.clip(g, 0, 1)
            warnings.warn(f"Green color outside range 0 - 1. Clipping to {g}")
        if b < 0 or b > 1:
            b = np.clip(b, 0, 1)
            warnings.warn(f"Blue color outside range 0 - 1. Clipping to {b}")
        request = roivolume_pb2.ROIVolumeSetColorRequest(roi=self.pb2_object, r=r, g=g, b=b)
        response = self.osirix_service_stub.ROIVolumeSetColor(request)
        self.response_check(response)

    @property
    @pyosirix_connection_check
    def opacity(self) -> float:
        """ The opacity of the ROI volume in the range 0-1
        """
        response = self.osirix_service_stub.ROIVolumeOpacity(self.pb2_object)
        self.response_check(response)
        return float(response.opacity)

    @opacity.setter
    @pyosirix_connection_check
    def opacity(self, opacity: float):
        """ The opacity of the ROI volume in the range 0-1
        """
        if opacity < 0 or opacity > 1:
            opacity = np.clip(opacity, 0, 1)
            warnings.warn(f"Opacity outside range 0 - 1. Clipping to {opacity}")
        request = roivolume_pb2.ROIVolumeSetOpacityRequest(roi=self.pb2_object,
                                                           opacity=opacity)
        response = self.osirix_service_stub.ROIVolumeSetOpacity(request)
        self.response_check(response)

    @property
    @pyosirix_connection_check
    def factor(self) -> float:
        """ The factor of the ROI volume in the range 0-1
        """
        response = self.osirix_service_stub.ROIVolumeFactor(self.pb2_object)
        self.response_check(response)
        return float(response.factor)

    @factor.setter
    @pyosirix_connection_check
    def factor(self, factor: float):
        """ The factor of the ROI volume in the range 0-1
        """
        if factor < 0 or factor > 1:
            factor = np.clip(factor, 0, 1)
            warnings.warn(f"Factor outside range 0 - 1. Clipping to {factor}")
        request = roivolume_pb2.ROIVolumeSetFactorRequest(roi=self.pb2_object,
                                                          factor=factor)
        response = self.osirix_service_stub.ROIVolumeSetFactor(request)
        self.response_check(response)

    @pyosirix_connection_check
    def volume(self) -> float:
        """ The volume of the ROI

        Returns:
            The volume of the 3D ROI (in millilitres).
        """
        response = self.osirix_service_stub.ROIVolumeVolume(self.pb2_object)
        self.response_check(response)
        return float(response.volume)


class ROI(osirix.base.OsirixBase):
    """ Represents a single region of interest in a 2D viewer (`ViewerController`)

    """
    def __repr__(self):
        return f"ROI: " \
               f"{self.name}"

    @classmethod
    def itypes(cls, reverse_order: bool = False) -> Dict:
        """ A map between the integer ROI type (itype) and a text descriptor.

        Args:
            reverse_order (bool): If `False` then the mapping is provided in the form
                `{itype: descriptor}` else it is returned in reverse order: `{descriptor: itype}`.

        Returns:
            A mapping between ROI itype and textual descriptor.
        """
        mapping = {5: "tMeasure",
                   6: "tROI",
                   9: "tOval",
                   10: "tOPolygon",
                   11: "tCPolygon",
                   12: "tAngle",
                   13: "tText",
                   14: "tArrow",
                   15: "tPencil",
                   19: "t2DPoint",
                   20: "tPlain",
                   26: "tAxis",
                   27: "tDynAngle",
                   29: "tTAGT",
                   30: "tBall",
                   31: "tOvalAngle"}
        if reverse_order:
            mapping = {value: key for key, value in mapping.items()}
        return mapping

    @property
    @pyosirix_connection_check
    def color(self) -> Tuple[int, int, int]:
        """ The color of the volume ROI as a (r, g, b) tuple (each channel in range 0-255)
        """
        response = self.osirix_service_stub.ROIColor(self.pb2_object)
        self.response_check(response)
        return response.r, response.g, response.b

    @color.setter
    @pyosirix_connection_check
    def color(self, color: Tuple[int, int, int]) -> None:
        """ The color of the volume ROI as a (r, g, b) tuple (each channel in range 0-255)
        """
        r, g, b = color
        if r < 0 or r > 255:
            r = np.clip(r, 0, 255)
            warnings.warn(f"Red color outside range 0 - 255. Clipping to {r}")
        if g < 0 or g > 255:
            g = np.clip(g, 0, 255)
            warnings.warn(f"Green color outside range 0 - 255. Clipping to {g}")
        if b < 0 or b > 255:
            b = np.clip(b, 0, 255)
            warnings.warn(f"Blue color outside range 0 - 255. Clipping to {b}")
        request = roi_pb2.ROISetColorRequest(roi=self.pb2_object, r=r, g=g, b=b)
        response = self.osirix_service_stub.ROISetColor(request)
        self.response_check(response)

    @property
    @pyosirix_connection_check
    def name(self) -> str:
        """ The name of the ROI
        """
        response = self.osirix_service_stub.ROIName(self.pb2_object)
        self.response_check(response)
        return response.name

    @name.setter
    @pyosirix_connection_check
    def name(self, name: str) -> None:
        """ The name of the ROI
        """
        request = roi_pb2.ROISetNameRequest(roi=self.pb2_object, name=name)
        response = self.osirix_service_stub.ROISetName(request)
        self.response_check(response)

    @property
    @pyosirix_connection_check
    def opacity(self) -> float:
        """ The opacity of the ROI, ranging from 0 (transparent) to 1 (opaque)
        """
        response = self.osirix_service_stub.ROIOpacity(self.pb2_object)
        self.response_check(response)
        return response.opacity

    @opacity.setter
    @pyosirix_connection_check
    def opacity(self, opacity: float) -> None:
        """ The opacity of the ROI, ranging from 0 (transparent) to 1 (opaque)
        """
        if opacity < 0 or opacity > 1:
            opacity = np.clip(opacity, 0, 1)
            warnings.warn(f"Opacity outside range 0 - 1. Clipping to {opacity}")
        request = roi_pb2.ROISetOpacityRequest(roi=self.pb2_object, opacity=opacity)
        response = self.osirix_service_stub.ROISetOpacity(request)
        self.response_check(response)

    @property
    @pyosirix_connection_check
    def points(self) -> NDArray:
        """ The vertices of any polygon-based ROI. An array with shape [N, 2].
        """
        response = self.osirix_service_stub.ROIPoints(self.pb2_object)
        self.response_check(response)
        points = []
        for i in range(len(response.points)):
            points.append([response.points[i].x, response.points[i].y])
        return np.array(points)

    @points.setter
    @pyosirix_connection_check
    def points(self, points: NDArray) -> None:
        """ The vertices of any polygon-based ROI. An array with shape [N, 2].
        """
        if points.ndim != 2:
            raise ValueError("Input points must have shape [N, 2]")
        if points.shape[1] != 2:
            raise ValueError("Input points must have shape [N, 2]")
        requested_points = []
        for i in range(points.shape[0]):
            requested_points.append(roi_pb2.ROISetPointsRequest.Point2D(x=points[i, 0],
                                                                        y=points[i, 1]))
        request = roi_pb2.ROISetPointsRequest(roi=self.pb2_object, points=requested_points)
        response = self.osirix_service_stub.ROISetPoints(request)
        self.response_check(response)

    @property
    @pyosirix_connection_check
    def thickness(self) -> float:
        """ The line thickness (in points) of any polygon-based ROI.
        """
        response = self.osirix_service_stub.ROIThickness(self.pb2_object)
        self.response_check(response)
        return response.thickness

    @thickness.setter
    @pyosirix_connection_check
    def thickness(self, thickness: float) -> None:
        """ The line thickness (in points) of any polygon-based ROI.
        """
        if thickness < 0:
            thickness = 0
            warnings.warn(f"Thickness < 0. Clipping to {thickness}")
        request = roi_pb2.ROISetThicknessRequest(roi=self.pb2_object, thickness=thickness)
        response = self.osirix_service_stub.ROISetThickness(request)
        self.response_check(response)

    @property
    @pyosirix_connection_check
    def pix(self) -> osirix.dcm_pix.DCMPix:
        """ The DCMPix instance coupled with the ROI. This should be the one on which it is drawn.
        """
        response = self.osirix_service_stub.ROIPix(self.pb2_object)
        self.response_check(response)
        return osirix.dcm_pix.DCMPix(self.osirix_service, response.pix)

    @property
    @pyosirix_connection_check
    def itype(self) -> int:
        # TODO: We need to more than provide an integer here. It should be a text description.
        """ The ROI type as an integer
        """
        response = self.osirix_service_stub.ROIIType(self.pb2_object)
        self.response_check(response)
        return int(response.itype)

    @property
    @pyosirix_connection_check
    def stype(self) -> str:
        """ The type of the ROI as a string """
        return self.itypes()[self.itype]

    @pyosirix_connection_check
    def delete(self) -> None:
        """ Delete the ROI.  This cannot be undone!
        """
        response = self.osirix_service_stub.ROIDelete(self.pb2_object)
        self.response_check(response)

    @pyosirix_connection_check
    def centroid(self) -> Tuple[float, float]:
        """ The centroid of the ROI in format (column, row). Non-integer values are permitted.
        """
        response = self.osirix_service_stub.ROICentroid(self.pb2_object)
        self.response_check(response)
        return response.x, response.y

    @pyosirix_connection_check
    def flip_horizontally(self) -> None:
        """ Flip the ROI horizontally
        """
        response = self.osirix_service_stub.ROIFlipHorizontally(self.pb2_object)
        self.response_check(response)

    @pyosirix_connection_check
    def flip_vertically(self) -> None:
        """ Flip the ROI vertically
        """
        response = self.osirix_service_stub.ROIFlipVertically(self.pb2_object)
        self.response_check(response)

    @pyosirix_connection_check
    def roi_area(self) -> float:
        """ The area of the ROI in squared centimeters.
        """
        response = self.osirix_service_stub.ROIArea(self.pb2_object)
        self.response_check(response)
        return response.area

    @pyosirix_connection_check
    def roi_move(self, columns: int, rows: int) -> None:
        # TODO: This could be float, but need to change the protocol definition.
        """ Move the ROI by a specified number of columns and rows.

        Args:
            columns (int): The number of columns to move.
                positive = left &rarr; right, negative = right &rarr; left.
            rows (int): The number of rows to move.
                positive = downwards, negative = upwards.
        """
        request = roi_pb2.ROIMoveRequest(roi=self.pb2_object, columns=columns, rows=rows)
        response = self.osirix_service_stub.ROIMove(request)
        self.response_check(response)

    @pyosirix_connection_check
    def rotate(self, theta: float, center: Tuple[float, float] = None) -> None:
        """ Rotate the ROI by angle theta, about a given

        Args:
            theta (float): The angle by which to rotate (in degrees). Positive is clockwise.
            center (Tuple[float, float]): The position (column, row) about which to rotate.
                These can be non-integer values. If `None`, then rotation is about the centroid of
                the ROI.
        """
        if center is None:
            center = self.centroid()
        column, row = center
        request = roi_pb2.ROIRotateRequest(roi=self.pb2_object, degrees=theta, x=column, y=row)
        response = self.osirix_service_stub.ROIRotate(request)
        self.response_check(response)



""" Provide access to the core OsiriX objects: 2D/3D viewers and the database (browser).

Example usage:
    ```python
    import osirix

    port = 50051    # You can use any unused port
    domain = "localhost"
    osirix_service = osirix.osirix_utilities.OsirixService(domain=domain, port=port)
    ```
"""

from __future__ import annotations
from typing import List, Tuple, Dict

import grpc  # noqa

from osirixgrpc import osirix_pb2_grpc, types_pb2
import osirixgrpc.osirix_pb2 as osirix_pb2
import osirixgrpc.utilities_pb2 as utilities_pb2

import osirix
from osirix.base import pyosirix_connection_check


class OsirixService(object):
    """ Create and manage an OsiriX gRPC service (stub).

    Attributes:
        domain (str): The domain with which to establish the connection.  Currently, only
            "127.0.0.1" is supported (localhost). Default is "127.0.0.1".
        port (int): The port number with which to establish the connection. Default is 50051.
        max_send_message_length (int): The maximum number of bytes permitted in a send message.
            Default is 500000000 (500 MB).
        max_receive_message_length (int): The maximum number of bytes permitted in a receive
            message. Default is 500000000 (500 MB).
        osirix_service_stub (osirix_pb2_grpc.OsiriXServiceStub): The service stub used to send gRPC
            messages to the OsiriXgrpc server. None if no connection established.
        channel (grpc._channel.Channel): An insecure gRPC channel configuration.
    """
    def __init__(self, domain: str = "127.0.0.1", port: int = 50051,
                 max_send_message_length: int = 512 * 1024 * 1024,
                 max_receive_message_length: int = 512 * 1024 * 1024):
        self.port = port
        self.domain = domain
        self.server_url = domain + ":" + str(self.port)
        self.max_send_message_length = max_send_message_length
        self.max_receive_message_length = max_receive_message_length
        self.osirix_service_stub = None
        self.channel = None

    def start_service(self):
        """ Start the insecure client service.

        Raises:
            GrpcException: Occurs when something goes wrong trying to set up the connection.

        """
        self.channel = grpc.insecure_channel(self.server_url,
                                             options=[("grpc.max_receive_message_length",
                                                       self.max_receive_message_length),
                                                      ("grpc.max_send_message_length",
                                                       self.max_send_message_length)])
        self.osirix_service_stub = osirix_pb2_grpc.OsiriXServiceStub(self.channel)
        if not self.check_connection():
            raise osirix.exceptions.GrpcException("Could not establish a connection with OsiriX.")

    def check_connection(self) -> bool:
        """ Check that a connection with OsiriX is established.

        Returns:
            True if a connection is established.
        """
        try:
            self.osirix_service_stub.OsirixVersion(utilities_pb2.Empty())
        except grpc.RpcError:
            return False
        return True

    def stop_service(self):
        """ Stop the insecure client service. """
        if self.channel:
            self.channel.close()


class Osirix(osirix.base.OsirixBase):
    """ Provides access functionality to core OsiriX objects (2D/3D viewers and Dicom browser).

    Note that it is better not to use this class directly, unless you know what you are doing.
    There is a global instance of this class, set-up according to the OsiriXgrpc plugin
    configuration, as described in the main `__init__.py` module.
    """

    @pyosirix_connection_check
    def current_browser(self) -> osirix.browser_controller.BrowserController:
        """ Return an instance of the current Dicom database browser.

        Note that it is better not use this method directly. Use the following instead:

        ```python
        import osirix
        browser = osirix.current_browser()
        ```

        Returns:
            The main Dicom browser instance.
        """
        request = utilities_pb2.Empty()
        response = self.osirix_service_stub.OsirixCurrentBrowser(request)
        self.response_check(response)
        browser_controller = response.browser_controller
        browser_controller_obj = osirix.browser_controller.BrowserController(self.osirix_service,
                                                                             browser_controller)
        return browser_controller_obj

    @pyosirix_connection_check
    def frontmost_viewer(self) -> osirix.viewer_controller.ViewerController:
        """ Return an instance of the front-most 2D viewer.

        If there are more than one, look for the viewer highlighted by a red boundary.
        Note that it is better not use this method directly. Use the following instead:

        ```python
        import osirix
        viewer = osirix.frontmost_viewer()
        ```

        Returns:
            The front-most 2D viewer.
        """
        request = utilities_pb2.Empty()
        response = self.osirix_service_stub.OsirixFrontmostViewer(request)
        self.response_check(response)
        viewer_controller = response.viewer_controller
        viewer_controller_obj = osirix.viewer_controller.ViewerController(self.osirix_service,
                                                                          viewer_controller)
        return viewer_controller_obj

    @pyosirix_connection_check
    def displayed_2d_viewers(self) -> List[osirix.viewer_controller.ViewerController, ...]:
        """ Return all displayed 2D viewer instances.

        Note that it is better not use this method directly. Use the following instead:

        ```python
        import osirix
        viewers = osirix.displayed_2d_viewers()
        ```

        Returns:
            A list of 2D viewer instances.
        """
        request = utilities_pb2.Empty()
        response = self.osirix_service_stub.OsirixDisplayed2DViewers(request)
        self.response_check(response)
        viewer_controller_objs = []
        for viewer_controller in response.viewer_controllers:
            viewer_controller_obj = osirix.viewer_controller.ViewerController(self.osirix_service,
                                                                              viewer_controller)
            viewer_controller_objs.append(viewer_controller_obj)
        return viewer_controller_objs

    @pyosirix_connection_check
    def frontmost_vr_controller(self) -> osirix.vr_controller.VRController:
        """ Return an instance of the front-most 3D viewer.

        Note that it is better not use this method directly. Use the following instead:

        ```python
        import osirix
        vr_controller = osirix.frontmost_vr_controller()
        ```

        Returns:
            The front-most 3D viewer.
        """
        request = utilities_pb2.Empty()
        response = self.osirix_service_stub.OsirixFrontmostVRController(request)
        self.response_check(response)
        vr_controller = response.vr_controller
        vr_controller_obj = osirix.vr_controller.VRController(self.osirix_service,
                                                              vr_controller)
        return vr_controller_obj

    @pyosirix_connection_check
    def displayed_vr_controllers(self) -> List[osirix.vr_controller.VRController, ...]:
        """ Return all displayed 3D viewer instances.

        Note that it is better not use this method directly. Use the following instead:

        ```python
        import osirix
        vr_controllers = osirix.displayed_vr_controllers()
        ```

        Returns:
            A list of 3D viewer instances.
        """
        request = utilities_pb2.Empty()
        response = self.osirix_service_stub.OsirixDisplayedVRControllers(request)
        self.response_check(response)
        vr_controller_objs = []
        for vr_controller in response.vr_controllers:
            vr_controller_obj = osirix.vr_controller.VRController(self.osirix_service,
                                                                  vr_controller)
            vr_controller_objs.append(vr_controller_obj)
        return vr_controller_objs

    @pyosirix_connection_check
    def osirix_version(self) -> Tuple[str, str]:
        """ Return the current version of OsiriX being linked to.

        Note that it is better not use this method directly. Use the following instead:

        ```python
        import osirix
        version = osirix.osirix_version()
        ```

        Returns:
            str: The current OsiriX version.
            str: The bundle name (e.g. "OsiriX MD")
        """
        request = utilities_pb2.Empty()
        response = self.osirix_service_stub.OsirixVersion(request)
        self.response_check(response)
        return response.version, response.bundle_name

    def __osirix_cache_uids__(self) -> List[str]:
        """ Return a list of UIDs available to osirixgrpc

        Returns:
            List: The list of available UIDs.
        """
        response = self.osirix_service_stub.OsirixCacheUids(utilities_pb2.Empty())
        self.response_check(response)
        return response.uids

    def __osirix_cache_object_for_uid__(self, uid: str):
        """ Return an object for a particular UID

        Args:
            uid (str): The UID for which to obtain the object

        Returns:
            id: The object.
        """
        request = osirix_pb2.OsirixCacheObjectForUidRequest(uid=uid)
        response = self.osirix_service_stub.OsirixCacheObjectForUid(request)
        self.response_check(response)
        if response.object_type == "BrowserController":
            return osirix.browser_controller.BrowserController(self.osirix_service,
                                                               types_pb2.BrowserController(
                                                                   osirixrpc_uid=uid))
        elif response.object_type == "ViewerController":
            return osirix.viewer_controller.ViewerController(self.osirix_service,
                                                             types_pb2.ViewerController(
                                                                 osirixrpc_uid=uid))
        elif response.object_type == "VRController":
            return osirix.vr_controller.VRController(self.osirix_service,
                                                     types_pb2.VRController(osirixrpc_uid=uid))
        elif response.object_type == "DicomImage":
            return osirix.dicom.DicomImage(self.osirix_service,
                                           types_pb2.DicomImage(osirixrpc_uid=uid))
        elif response.object_type == "DicomSeries":
            return osirix.dicom.DicomSeries(self.osirix_service,
                                           types_pb2.DicomSeries(osirixrpc_uid=uid))
        elif response.object_type == "DicomStudy":
            return osirix.dicom.DicomStudy(self.osirix_service,
                                           types_pb2.DicomStudy(osirixrpc_uid=uid))
        elif response.object_type == "ROI":
            return osirix.roi.ROI(self.osirix_service, types_pb2.ROI(osirixrpc_uid=uid))
        elif response.object_type == "ROIVolume":
            return osirix.roi.ROIVolume(self.osirix_service, types_pb2.ROIVolume(osirixrpc_uid=uid))
        elif response.object_type == "DCMPix":
            return osirix.dcm_pix.DCMPix(self.osirix_service, types_pb2.DCMPix(osirixrpc_uid=uid))
        else:
            return response.object_type

    def __osirix_cache_objects__(self) -> Dict:
        """ Return an object for a particular UID

        Returns:
            Dict: All objects keyed by the uid: {uid: object}
        """
        uids = self.__osirix_cache_uids__()
        return {uid: self.__osirix_cache_object_for_uid__(uid) for uid in uids}



""" Exceptions classes for pyOsiriX.

"""


class GrpcException(Exception):
    """Exception raised when connection to OsiriX cannot be established.

    Attributes:
        message (str): An explanation of the error
    """
    def __init__(self, message: str = ""):
        self.message = message
        super().__init__(self.message)



""" Functionality for the 3D viewer.

Example usage:
    ```python
    import osirix

    vr_controller = osirix.frontmost_vr_controller()  # Raises GrpcException if not available.
    vr_controller.wlww = (50, 100)  # Set the window level and width
    print(f"This window has {len(vr_controller.roi_volumes())} ROI volumes")
    ```
"""

from __future__ import annotations
from typing import Tuple, List

import osirixgrpc.vrcontroller_pb2 as vrcontroller_pb2

import osirix
from osirix.base import pyosirix_connection_check


class VRController(osirix.base.OsirixBase):
    """ Represents one of the displayed volume render windows.

    """
    def __repr__(self):
        return f"VRController: " \
               f"{self.title} " \
               f"({self.rendering_mode})"

    @property
    @pyosirix_connection_check
    def rendering_mode(self) -> str:
        """ The rendering mode: "VR" for volume render, "MIP" for maximum intensity projection.
        """
        response = self.osirix_service_stub.VRControllerRenderingMode(self.pb2_object)
        self.response_check(response)
        return response.rendering_mode

    @rendering_mode.setter
    @pyosirix_connection_check
    def rendering_mode(self, rendering_mode: str) -> None:
        """ The rendering mode: "VR" for volume render, "MIP" for maximum intensity projection.
        """
        if rendering_mode not in ["VR", "MIP"]:
            raise ValueError("Bad rendering_mode. Must be 'VR' or 'MIP'")
        request = vrcontroller_pb2.VRControllerSetRenderingModeRequest(
            vr_controller=self.pb2_object, rendering_mode=rendering_mode)
        response = self.osirix_service_stub.VRControllerSetRenderingMode(request)
        self.response_check(response)

    @property
    @pyosirix_connection_check
    def style(self) -> str:
        """ The style of the volume render window.
        """
        response = self.osirix_service_stub.VRControllerStyle(self.pb2_object)
        self.response_check(response)
        return response.style

    @property
    @pyosirix_connection_check
    def title(self) -> str:
        """ The title of the volume render window.
        """
        response = self.osirix_service_stub.VRControllerTitle(self.pb2_object)
        self.response_check(response)
        return response.title

    @property
    @pyosirix_connection_check
    def wlww(self) -> Tuple[float, float]:
        """ The window level and window width of the viewer.
        """
        response = self.osirix_service_stub.VRControllerWLWW(self.pb2_object)
        self.response_check(response)
        return response.wl, response.ww

    @wlww.setter
    @pyosirix_connection_check
    def wlww(self, wlww: Tuple[float, float]) -> None:
        """The window level and window width of the viewer.
        """
        request = vrcontroller_pb2.VRControllerSetWLWWRequest(vr_controller=self.pb2_object,
                                                              wl=wlww[0],
                                                              ww=wlww[1])
        response = self.osirix_service_stub.VRControllerSetWLWW(request)
        self.response_check(response)

    @pyosirix_connection_check
    def blending_controller(self) -> osirix.viewer_controller.ViewerController:
        """ The 2D ViewerController instance currently being blended (fused).

        Returns:
            The fused ViewerController instance.

        Example usage:
            ```python
            vr_controllers = frontmost_viewer.vr_controllers()
            if len(vr_controllers) > 0:
                vr_controller = vr_controllers[0]
                fused_viewer = vr_controller.blending_controller()
            ```
        """
        response = self.osirix_service_stub.VRControllerBlendingController(self.pb2_object)
        self.response_check(response)
        return osirix.viewer_controller.ViewerController(self.osirix_service,
                                                         response.viewer_controller)

    @pyosirix_connection_check
    def viewer_2d(self) -> osirix.viewer_controller.ViewerController:
        """ The 2D ViewerController instance from which the 3D viewer was started.

        Returns:
            The ViewerController instance.

        Example usage:
            ```python
            vr_controllers = frontmost_viewer.vr_controllers()
            if len(vr_controllers) > 0:
                vr_controller = vr_controllers[0]
                viewer_controller = vr_controller.viewer_2d()
            ```
        """
        response = self.osirix_service_stub.VRControllerViewer2D(self.pb2_object)
        self.response_check(response)
        return osirix.viewer_controller.ViewerController(self.osirix_service,
                                                         response.viewer_controller)

    @pyosirix_connection_check
    def roi_volumes(self) -> List[osirix.roi.ROIVolume]:
        """ Obtain references to all volumetric ROIs.

        Returns:
            A list of ROIVolume instances.

        Example usage:
            ```python
            frontmost_viewer = osirix.frontmost_viewer()
            vr_controllers = frontmost_viewer.vr_controllers()
            if len(vr_controllers) == 0:
                print("No open VR Controller.  Creating one.")
                vr_controller = frontmost_viewer.open_vr_viewer(mode="MIP")
            else:
                vr_controller = vr_controllers[0]

            roi_volumes = vr_controller.roi_volumes()
            if len(roi_volumes) == 0:
                print("No valid ROI volumes available")
            for roi_volume in roi_volumes:
                print(f"Displaying ROI volume with name {roi_volume.name}")
                roi_volume.visible = True
            ```
        """
        response = self.osirix_service_stub.VRControllerROIVolumes(self.pb2_object)
        self.response_check(response)
        roi_volumes = []
        for roi_volume in response.roi_volumes:
            roi_volumes.append(osirix.roi.ROIVolume(self, self.osirix_service, roi_volume))
        return roi_volumes

    @pyosirix_connection_check
    def hide_roi_volume(self, roi_volume: osirix.roi.ROIVolume):
        """ Ensure that an input ROI is hidden on the 3D display.

        Args:
            roi_volume (osirix.roi.ROIVolume): The ROI volume to hide.
        """
        request = vrcontroller_pb2.VRControllerHideROIVolumeRequest(
            vr_controller=self.pb2_object, roi_volume=roi_volume.pb2_object)
        response = self.osirix_service_stub.VRControllerHideROIVolume(request)
        self.response_check(response)

    @pyosirix_connection_check
    def display_roi_volume(self, roi_volume: osirix.roi.ROIVolume):
        """ Ensure that an input ROI is visible on the 3D display.

        Args:
            roi_volume (osirix.roi.ROIVolume): The ROI volume to make visible.
        """
        request = vrcontroller_pb2.VRControllerDisplayROIVolumeRequest(
            vr_controller=self.pb2_object, roi_volume=roi_volume.pb2_object)
        response = self.osirix_service_stub.VRControllerDisplayROIVolume(request)
        self.response_check(response)

    @pyosirix_connection_check
    def needs_display_update(self):
        """ Update the VRController display

        Example usage:
            ```python
            vr_controllers = frontmost_viewer.vr_controllers()
            vr_controller = vr_controllers[0]
            vr_controller.needs_display_update()
            ```
        """
        response = self.osirix_service_stub.VRControllerNeedsDisplayUpdate(self.pb2_object)
        self.response_check(response)



""" Functionality for the 2D viewer.

Example usage:
    ```python
    import osirix

    viewer = osirix.frontmost_viewer()  # Raises GrpcException error if no viewer is available.
    n_slices = len(viewer.pix_list(0))  # pix_list(n) returns array of DCMPix instances for nth frame.
    viewer.idx = int(n_slices / 2)  # Move display to the central slice.
    viewer.wlww = (50, 100)  # Set the displayed window level and width..
    viewer.needs_display_update()  #Get the viewer to re-draw itself.
    ```
"""

from __future__ import annotations
from typing import Tuple, List
import warnings

from numpy.typing import NDArray
import numpy as np

import osirixgrpc.viewercontroller_pb2 as viewercontroller_pb2

import osirix  # noqa
from osirix.base import pyosirix_connection_check  # noqa


class ViewerController(osirix.base.OsirixBase):
    """ Represents one of the displayed volume render windows.

    """
    def __repr__(self):
        return f"ViewerController: " \
               f"{self.title}"

    @property
    @pyosirix_connection_check
    def idx(self) -> int:
        """ The slice index currently being displayed to the viewer (starting at 0).

        Note that the index displayed on the viewer is this value plus one.
        """
        response = self.osirix_service_stub.ViewerControllerIdx(self.pb2_object)
        self.response_check(response)
        return int(response.idx)

    @idx.setter
    @pyosirix_connection_check
    def idx(self, idx: int) -> None:
        """ The data slice index currently being displayed to the viewer (starting at 0).

        Note that this may not be equal to the index displayed on the viewer when `data_flipped` is
            True.
        """
        request = viewercontroller_pb2.ViewerControllerSetIdxRequest(
            viewer_controller=self.pb2_object, idx=idx)
        response = self.osirix_service_stub.ViewerControllerSetIdx(request)
        self.response_check(response)

    @property
    @pyosirix_connection_check
    def displayed_idx(self) -> int:
        """ The data slice index currently being displayed to the viewer (starting at 0).

        Note that this may not be equal to the index displayed on the viewer when `data_flipped` is
            True.
        """
        response = self.osirix_service_stub.ViewerControllerDisplayedIdx(self.pb2_object)
        self.response_check(response)
        return int(response.displayed_idx)

    @displayed_idx.setter
    @pyosirix_connection_check
    def displayed_idx(self, idx: int) -> None:
        """ The data slice index currently being displayed to the viewer (starting at 0).

        Note that the index displayed on the viewer is this value plus one.
        """
        request = viewercontroller_pb2.ViewerControllerSetDisplayedIdxRequest(
            viewer_controller=self.pb2_object, displayed_idx=idx)
        response = self.osirix_service_stub.ViewerControllerSetDisplayedIdx(request)
        self.response_check(response)

    @property
    @pyosirix_connection_check
    def flipped_data(self) -> int:
        """ Whether the slices data are flipped in the viewer.
        """
        response = self.osirix_service_stub.ViewerControllerFlippedData(self.pb2_object)
        self.response_check(response)
        return int(response.flipped_data)

    @property
    @pyosirix_connection_check
    def modality(self) -> str:
        """ The modality of the displayed data.
        """
        response = self.osirix_service_stub.ViewerControllerModality(self.pb2_object)
        self.response_check(response)
        return response.modality

    @property
    @pyosirix_connection_check
    def movie_idx(self) -> int:
        """ The frame currently being displayed by the viewer.
        """
        response = self.osirix_service_stub.ViewerControllerMovieIdx(self.pb2_object)
        self.response_check(response)
        return response.movie_idx

    @movie_idx.setter
    @pyosirix_connection_check
    def movie_idx(self, movie_idx: int) -> None:
        """ The frame currently being displayed by the viewer.
        """
        if movie_idx < 0:
            raise ValueError("movie_idx must be >= 0.")
        request = viewercontroller_pb2.ViewerControllerSetMovieIdxRequest(
            viewer_controller=self.pb2_object, movie_idx=movie_idx)
        response = self.osirix_service_stub.ViewerControllerSetMovieIdx(request)
        self.response_check(response)

    @property
    @pyosirix_connection_check
    def title(self) -> str:
        """ The title of the 2D viewing window.
        """
        response = self.osirix_service_stub.ViewerControllerTitle(self.pb2_object)
        self.response_check(response)
        return response.title

    @property
    @pyosirix_connection_check
    def wlww(self) -> Tuple[float, float]:
        """The window level and window width of the viewer.
        """
        response = self.osirix_service_stub.ViewerControllerWLWW(self.pb2_object)
        self.response_check(response)
        return response.wl, response.ww

    @wlww.setter
    @pyosirix_connection_check
    def wlww(self, wlww: Tuple[float, float]) -> None:
        """The window level and window width of the viewer.
        """
        request = viewercontroller_pb2.ViewerControllerSetWLWWRequest(
            viewer_controller=self.pb2_object, wl=wlww[0], ww=wlww[1])
        response = self.osirix_service_stub.ViewerControllerSetWLWW(request)
        self.response_check(response)

    @property
    @pyosirix_connection_check
    def max_movie_index(self) -> int:
        """ The maximum movie index in the viewer.
        """
        response = self.osirix_service_stub.ViewerControllerMaxMovieIdx(self.pb2_object)
        self.response_check(response)
        return response.max_movie_idx

    @pyosirix_connection_check
    def close(self):
        """ Close the viewer.

        Note: This will close any associated viewers (e.g. VRControllers).
        """
        response = self.osirix_service_stub.ViewerControllerCloseViewer(self.pb2_object)
        self.response_check(response)

    @pyosirix_connection_check
    def pix_list(self, movie_idx: int = None) -> List[osirix.dcm_pix.DCMPix]:
        """ Access the list of DCMPix objects contained within the 2D viewer.

        Args:
            movie_idx (int): The frame from which to extract the DCMPix instances. If `None`, then
                use the currently displayed movie_idx.

        Returns:
            The list of osirix.dcm_pix.DCMPix instances requested for a particular frame.
        """
        if movie_idx is None:
            movie_idx = self.movie_idx
        request = viewercontroller_pb2.ViewerControllerPixListRequest(
            viewer_controller=self.pb2_object, movie_idx=movie_idx)
        response = self.osirix_service_stub.ViewerControllerPixList(request)
        self.response_check(response)
        return [osirix.dcm_pix.DCMPix(self.osirix_service, pix) for pix in response.pix]

    @pyosirix_connection_check
    def roi_list(self, movie_idx: int = None) -> List[List[osirix.roi.ROI]]:
        """ Access a list of lists of ROI objects contained within the 2D viewer.

        The length of the outer list is the same as the number of slices in the viewer. Note
        that some inner lists will be empty if no ROIs are drawn on that slice.

        Args:
            movie_idx (int): The frame from which to extract the ROI instances. If `None`, then
                use the currently displayed movie_idx.

        Returns:
            A list of lists of osirix.roi.ROI instances requested for a particular frame.
        """
        if movie_idx is None:
            movie_idx = self.movie_idx

        # Get the response
        request = viewercontroller_pb2.ViewerControllerROIListRequest(
            viewer_controller=self.pb2_object, movie_idx=movie_idx)
        response = self.osirix_service_stub.ViewerControllerROIList(request)
        self.response_check(response)

        # Loop through all ROI slices and append a list of inner ROIs.
        rois = []
        for roi_slice in response.roi_slices:
            rois_ = []
            for roi in roi_slice.rois:
                rois_.append(osirix.roi.ROI(self.osirix_service, roi))
            rois.append(rois_)
        return rois

    @pyosirix_connection_check
    def rois_with_name(self, name: str, movie_idx: int = None, in_4d: bool = False)\
            -> List[osirix.roi.ROI]:
        """ Return a list of ROIs with a given name.
        Args:
            name (str): The name of ROIs to look for.
            movie_idx (int): The frame from which to extract the ROI instances. If `None`, then
                use the currently displayed movie_idx.
            in_4d (bool): Whether to look through all frames.

        Returns:
            A list of ROIs with the given name.
        """
        if movie_idx is None:
            movie_idx = self.movie_idx

        request = viewercontroller_pb2.ViewerControllerROIsWithNameRequest(
            viewer_controller=self.pb2_object, name=name, movie_idx=movie_idx, in_4d=in_4d)
        response = self.osirix_service_stub.ViewerControllerROIsWithName(request)
        self.response_check(response)

        rois = []
        for roi in response.rois:
            rois.append(osirix.roi.ROI(self.osirix_service, roi))
        return rois

    @pyosirix_connection_check
    def selected_rois(self) -> List[osirix.roi.ROI]:
        """ Return a list of the user-selected ROIs

        Returns:
            A list of selected ROIs.
        """
        response = self.osirix_service_stub.ViewerControllerSelectedROIs(self.pb2_object)
        self.response_check(response)
        rois = []
        for roi in response.rois:
            rois.append(osirix.roi.ROI(self.osirix_service, roi))
        return rois

    @pyosirix_connection_check
    def vr_controllers(self) -> List[osirix.vr_controller.VRController]:
        """ Return the list of currently displayed 3D viewers associates with this 2D window.

        Returns:
            Any VRController instances associated with this 2D window.

        Example usage:
            ```python
            frontmost_viewer = osirix.frontmost_viewer()
            vr_controllers = frontmost_viewer.vr_controllers()
            ```
        """
        response = self.osirix_service_stub.ViewerControllerVRControllers(self.pb2_object)
        self.response_check(response)
        vrs = []
        for vr_controller in response.vr_controllers:
            vrs.append(osirix.vr_controller.VRController(self.osirix_service, vr_controller))
        return vrs

    @pyosirix_connection_check
    def blending_controller(self) -> ViewerController:
        """ Return the viewer controller instance that is being fused with this one.

        Returns:
            The blending viewer instance.

        Example usage:
            ```python
            frontmost_viewer = osirix.frontmost_viewer()
            blending_controller = frontmost_viewer.blending_controller()
            ```
        """
        response = self.osirix_service_stub.ViewerControllerBlendingController(self.pb2_object)
        self.response_check(response)
        return ViewerController(self.osirix_service, response.blending_viewer)

    @pyosirix_connection_check
    def fuse_with_viewer(self, viewer: ViewerController):
        """ Fuse another viewer onto this one.

        Args:
            viewer (osirix.viewer_controller.ViewerController): The viewer to fuse onto the present
                one.

        Example usage:
            ```python
            frontmost_viewer = osirix.frontmost_viewer()
            frontmost_viewer.fuse_with_viewer(another_viewer_controller)
            ```
        """
        request = viewercontroller_pb2.ViewerControllerFuseWithViewerRequest(
            viewer_controller=self.pb2_object, fusion_viewer_controller=viewer.pb2_object)
        response = self.osirix_service_stub.ViewerControllerFuseWithViewer(request)
        self.response_check(response)

    @pyosirix_connection_check
    def close_viewer(self) -> None:
        """ Close the viewer
        """
        response = self.osirix_service_stub.ViewerControllerCloseViewer(self.pb2_object)
        self.response_check(response)

    @pyosirix_connection_check
    def copy_viewer_window(self) -> ViewerController:
        """ Create a copy of the viewer.

        This can be useful for image processing when you want to keep a copy of the original and
        a processed version.

        Returns:
            The new osirix.viewer_controller.ViewerController instance.

        Example usage:
            ```python
            frontmost_viewer = osirix.frontmost_viewer()
            acopy = frontmost_viewer.copy_viewer_window()
            ```
        """
        response = self.osirix_service_stub.ViewerControllerCopyViewerWindow(self.pb2_object)
        self.response_check(response)
        return ViewerController(self.osirix_service, response.viewer_controller)

    @pyosirix_connection_check
    def cur_dcm(self) -> osirix.dcm_pix.DCMPix:
        """ Return the currently displayed DCMPix instance.

        Returns:
            The currently displayed DCMPix instance.

        Example usage:
            ```python
            frontmost_viewer = osirix.frontmost_viewer()
            pix = frontmost_viewer.cur_dcm()
            ```
        """
        response = self.osirix_service_stub.ViewerControllerCurDCM(self.pb2_object)
        self.response_check(response)
        return osirix.dcm_pix.DCMPix(self.osirix_service, response.pix)

    @pyosirix_connection_check
    def is_data_volumic(self, in_4d: bool = False) -> bool:
        """ Does the underlying data have sufficient geometric properties for volume rendering.

        Args:
            in_4d (bool): Determine for all frames.

        Returns:
            Is the data volumic?
        """
        request = viewercontroller_pb2.ViewerControllerIsDataVolumicRequest(
            viewer_controller=self.pb2_object, in_4d=in_4d)
        response = self.osirix_service_stub.ViewerControllerIsDataVolumic(request)
        self.response_check(response)
        return response.is_volumic

    @pyosirix_connection_check
    def needs_display_update(self) -> None:
        """ Tell the viewer it should update its display.

        This is particularly important if you want to see changes to pixel values in real time.
        """
        response = self.osirix_service_stub.ViewerControllerNeedsDisplayUpdate(self.pb2_object)
        self.response_check(response)

    @pyosirix_connection_check
    def resample_viewer_controller(self, vc: ViewerController) -> ViewerController:
        """ Resample this (moving) ViewerController based on another (fixed) ViewerController.

        Note that the original moving viewer is closed, so a reference to the new one is returned.
        This means it is generally advised to the call the method in the following manner:

        ```python
        moving_viewer = moving_viewer.resample_viewer_controller(fixed_viewer)
        ```

        Args:
            vc (osirix.viewer_controller.ViewerController): The fixed viewer.

        Returns:
            The resampled viewer window.
        """
        request = viewercontroller_pb2.ViewerControllerResampleViewerControllerRequest(
            viewer_controller=self.pb2_object, fixed_viewer_controller=vc.pb2_object)
        response = self.osirix_service_stub.ViewerControllerResampleViewerController(request)
        self.response_check(response)
        return ViewerController(self.osirix_service, response.resampled_viewer)

    @pyosirix_connection_check
    def open_vr_viewer(self, mode: str = "VR") -> osirix.vr_controller.VRController:
        """ Open a new 3D viewer for the enclosed data.

        Args:
            mode (str): Either 'VR' for volume render, or 'MIP' for a maximum intensity projection.

        Returns:
            A VRController instance.

        Example usage:
            ```python
            frontmost_viewer = osirix.frontmost_viewer()
            vr_controller = frontmost_viewer.open_vr_viewer(mode="VR")
            ```
        """
        if mode not in ["VR", "MIP"]:
            raise ValueError("`mode` must be either 'VR' or 'MIP'")
        request = viewercontroller_pb2.ViewerControllerOpenVRViewerForModeRequest(
            viewer_controller=self.pb2_object, mode=mode)
        response = self.osirix_service_stub.ViewerControllerOpenVRViewerForMode(request)
        self.response_check(response)
        return osirix.vr_controller.VRController(self.osirix_service, response.vr_controller)

    @pyosirix_connection_check
    def new_roi(self, itype: int = 15, name: str = "", idx: int = 0, movie_idx: int = 0,
                buffer_position_column: int = 0, buffer_position_row: int = 0,
                color: Tuple[float, float, float] = (0, 255, 0), thickness: float = 1.0,
                opacity: float = 1.0, buffer: NDArray = None, rect: NDArray = None,
                points: NDArray = None) -> osirix.roi.ROI:
        """ Create a new ROI within the viewer.

        __Important__: It is safer to create a new ROI using one of the utility functions as they
        provide better checking on input parameters in certain contexts, and ensure the right ROI
        type is created. This is meant as a parent function that performs the heavy lifting. It is
        made public for transparency.

        Args:
            itype (int): The index of the roi type (see `osirix.roi.ROI.itypes` for details).
            name (str): The name of the ROI.
            idx (int): The slice index on which to create the new ROI.
            movie_idx (int): The frame on which to create the new ROI.
            buffer_position_column (int): The column offset of a mask ROI in pixel coordinates.
            buffer_position_row (int): The row offset of a mask ROI in pixel coordinates.
            color (tuple): The RGB color of the ROI (values from 0-255 in each channel).
            thickness (float): The thickness of the ROI. Value between 1-20 as per OsiriX.
            opacity (float): The opacity on a scale from 0 (transparent) to 1 (opaque).
            buffer (NDArray): The buffer mask as a 2D boolean Numpy array. If `None`, then ignored.
            rect (NDArray): The rectangle defining the ROI. Must be shape (4,).
            points (NDArray): The points defining the ROI. Must be shape (N, 2) for N vertices. If
                `None` or `buffer` is not `None` or `rect` is not `None` then ignored.

        Returns:
             The created ROI instance.

        Raises:
            ValueError: When `buffer`, `points` and `rect` are all None.
            ValueError: When the input data is of invalid format.
        """
        if buffer is None and rect is None and points is None:
            raise ValueError("buffer, rect and points cannot be None at the same time.")

        if thickness < 0 or thickness > 20:
            thickness = np.clip(thickness, 0, 20)
            warnings.warn(f"Bad value for thickness. Clipping to {thickness}")

        color_request = viewercontroller_pb2.ViewerControllerNewROIRequest.Color(r=color[0],
                                                                                 g=color[1],
                                                                                 b=color[2])
        if itype == 20:
            if buffer is None:
                raise ValueError("Need buffer data if creating a tPlain ROI")
            buffer = np.array(buffer)
            if buffer.ndim != 2:
                raise ValueError("`buffer` must be two-dimensional")
            rows, columns = buffer.shape
            buffer_request = viewercontroller_pb2.ViewerControllerNewROIRequest.Buffer(
                buffer=buffer.ravel().tolist(), rows=rows, columns=columns)
            request = viewercontroller_pb2.ViewerControllerNewROIRequest(
                viewer_controller=self.pb2_object, idx=idx, movie_idx=movie_idx,
                buffer_position_x=buffer_position_column, buffer_position_y=buffer_position_row,
                color=color_request, opacity=opacity, name=name, buffer=buffer_request, itype=itype)
        elif itype in [6, 9, 13, 19, 31]:
            if rect is None:
                raise ValueError("Need rect if creating a rectangle-based ROI")
            rect = np.array(rect)
            if rect.ndim != 1:
                raise ValueError("`rect` must be one-dimensional")
            if rect.shape[0] != 4:
                raise ValueError("`rect` must be shape (4,)")
            rect_request = viewercontroller_pb2.ViewerControllerNewROIRequest.Rect(origin_x=rect[0],
                                                                                   origin_y=rect[1],
                                                                                   width=rect[2],
                                                                                   height=rect[3])
            request = viewercontroller_pb2.ViewerControllerNewROIRequest(
                viewer_controller=self.pb2_object, idx=idx, movie_idx=movie_idx,
                color=color_request, opacity=opacity, name=name, rectangle=rect_request,
                thickness=thickness, itype=itype)
        else:
            if points is None:
                raise ValueError("Need points if creating a point-based ROI")
            points = np.array(points)
            if points.ndim != 2:
                raise ValueError("`points` must be two-dimensional")
            points_request = [
                viewercontroller_pb2.ViewerControllerNewROIRequest.Point2D(x=p[0], y=p[1]) for p in
                points]
            request = viewercontroller_pb2.ViewerControllerNewROIRequest(
                viewer_controller=self.pb2_object, idx=idx, movie_idx=movie_idx,
                color=color_request, opacity=opacity, name=name, points=points_request,
                thickness=thickness, itype=itype)
        response = self.osirix_service_stub.ViewerControllerNewROI(request)
        self.response_check(response)
        return osirix.roi.ROI(self.osirix_service, response.roi)

    @pyosirix_connection_check
    def new_mask_roi(self, buffer: NDArray, **kwargs) -> osirix.roi.ROI:
        """ Create a new mask ROI within the viewer.

        Args:
            buffer (NDArray): The buffer mask as a 2D boolean Numpy array.
            **kwargs: See `new_roi` for all additional parameters.

        Returns:
             The created ROI instance.
        """
        return self.new_roi(buffer=buffer, itype=20, **kwargs)

    @pyosirix_connection_check
    def new_polygon_roi(self, points: NDArray, closed: bool = True, **kwargs) -> osirix.roi.ROI:
        """ Create a new polygon ROI within the viewer.

        Args:
            points (NDArray): The vertices of the ROI.  Must have shape (N, 2).
            closed (bool): Whether the polygon is closed.
            **kwargs: See `new_roi` for all additional parameters.

        Returns:
             The created ROI instance.
        """
        itype = 10
        if closed:
            itype = 11
        return self.new_roi(points=points, itype=itype, **kwargs)

    @pyosirix_connection_check
    def new_measurement_roi(self, start: Tuple[float, float], end: Tuple[float, float], **kwargs)\
            -> osirix.roi.ROI:
        """ Create a new length measurement ROI within the viewer.

        Args:
            start (float, float): The starting position of the ROI (column, row).
            end (float, float): The ending position of the ROI (column, row).
            **kwargs: See `new_roi` for all additional parameters.

        Returns:
             The created ROI instance.
        """
        points = np.array([[start[0], start[1]],
                           [end[0], end[1]]])
        return self.new_roi(points=points, itype=5, **kwargs)

    @pyosirix_connection_check
    def new_point_roi(self, column: float, row: float, **kwargs) -> osirix.roi.ROI:
        """ Create a new 2D point ROI within the viewer.

        Args:
            column (float): The column position of the ROI
            row (float): The row position of the ROI
            **kwargs: See `new_roi` for all additional parameters.

        Returns:
             The created ROI instance.
        """
        rect = np.array([column, row, 0, 0])
        return self.new_roi(rect=rect, itype=19, **kwargs)

    @pyosirix_connection_check
    def new_text_roi(self, column: float, row: float, name: str, **kwargs) -> osirix.roi.ROI:
        """ Create a new text-box ROI within the viewer.

        Args:
            column (float): The column position of the ROI
            row (float): The row position of the ROI
            name (str): The text to show
            **kwargs: See `new_roi` for all additional parameters.

        Returns:
             The created ROI instance.
        """
        rect = np.array([column, row, 0, 0])
        return self.new_roi(rect=rect, itype=13, name=name, **kwargs)

    @pyosirix_connection_check
    def new_arrow_roi(self, tail: Tuple[float, float], head: Tuple[float, float],  **kwargs)\
            -> osirix.roi.ROI:
        """ Create a new text-box ROI within the viewer.

        Args:
            tail (float, float): The (column, row) position of the arrow tail.
            head (float, float): The (column, row) position of the arrow hear.
            **kwargs: See `new_roi` for all additional parameters.

        Returns:
             The created ROI instance.
        """
        points = np.vstack([head, tail])
        return self.new_roi(points=points, itype=14, **kwargs)

    @pyosirix_connection_check
    def new_oval_roi(self, center: Tuple[float, float], width: float, height: float, **kwargs)\
            -> osirix.roi.ROI:
        """ Create a new oval ROI within the viewer.

        Args:
            center (float, float): The (column, row) center of the oval ROI.
            width (float): The width of the oval ROI.
            height (float): The height of the oval ROI.
            **kwargs: See `new_roi` for all additional parameters.

        Returns:
             The created ROI instance.
        """
        rect = np.array([center[0], center[1], width / 2, height / 2])
        return self.new_roi(rect=rect, itype=9, **kwargs)

    @pyosirix_connection_check
    def new_rectangle_roi(self, center: Tuple[float, float], width: float, height: float, **kwargs)\
            -> osirix.roi.ROI:
        """ Create a new rectangle ROI within the viewer.

        Args:
            center (float, float): The (column, row) center of the rectangle ROI.
            width (float): The width of the rectangle ROI.
            height (float): The height of the rectangle ROI.
            **kwargs: See `new_roi` for all additional parameters.

        Returns:
             The created ROI instance.
        """
        rect = np.array([center[0] - width / 2, center[1] - height / 2, width, height])
        return self.new_roi(rect=rect, itype=6, **kwargs)

    @pyosirix_connection_check
    def new_angle_roi(self, p1: Tuple[float, float], p2: Tuple[float, float],
                      p3: Tuple[float, float], **kwargs) -> osirix.roi.ROI:
        """ Create a new angle ROI within the viewer.

        This draws two connected vectors:

            - v = p2 &rarr; p1
            - u = p2 &rarr; p3

        The displayed angle is the angle between these two vectors.

        Args:
            p1 (float, float): The (column, row) position of the first angle vertex.
            p2 (float, float): The (column, row) position of the second angle vertex.
            p3 (float, float): The (column, row) position of the third angle vertex.
            **kwargs: See `new_roi` for all additional parameters.

        Returns:
             The created ROI instance.
        """
        points = np.vstack([np.array(p1), np.array(p2), np.array(p3)])
        return self.new_roi(points=points, itype=12, **kwargs)

    @pyosirix_connection_check
    def new_dynamic_angle_roi(self, p1: Tuple[float, float], p2: Tuple[float, float],
                              p3: Tuple[float, float], p4: Tuple[float, float], **kwargs)\
            -> osirix.roi.ROI:
        """ Create a new dynamic angle ROI within the viewer.

        This draws two vectors:

            - v = p2 &rarr; p1
            - u = p3 &rarr; p4

        The displayed angle is the angle between these two vectors. The vector p2 -> p3 is also
        drawn.

        Args:
            p1 (float, float): The (column, row) position of the first angle vertex.
            p2 (float, float): The (column, row) position of the second angle vertex.
            p3 (float, float): The (column, row) position of the third angle vertex.
            p4 (float, float): The (column, row) position of the fourth angle vertex.
            **kwargs: See `new_roi` for all additional parameters.

        Returns:
             The created ROI instance.
        """
        points = np.vstack([np.array(p1), np.array(p2), np.array(p3), np.array(p4)])
        return self.new_roi(points=points, itype=27, **kwargs)

    @pyosirix_connection_check
    def new_axis_roi(self, p1: Tuple[float, float], p2: Tuple[float, float],
                     p3: Tuple[float, float], p4: Tuple[float, float], **kwargs) -> osirix.roi.ROI:
        """ Create a new axis ROI within the viewer.

        Args:
            p1 (float, float): The (column, row) position of the first vertex.
            p2 (float, float): The (column, row) position of the second vertex.
            p3 (float, float): The (column, row) position of the third vertex.
            p4 (float, float): The (column, row) position of the fourth vertex.
            **kwargs: See `new_roi` for all additional parameters.

        Returns:
             The created ROI instance.
        """
        points = np.vstack([np.array(p1), np.array(p2), np.array(p3), np.array(p4)])
        return self.new_roi(points=points, itype=26, **kwargs)

    @pyosirix_connection_check
    def new_tagt_roi(self, a: Tuple[float, float], b: Tuple[float, float],
                     c: Tuple[float, float], d: Tuple[float, float], **kwargs) -> osirix.roi.ROI:
        """ Create a new tTAGT ROI within the viewer.

        ROI displays:

            - The length of vector DA.
            - The perpendicular distance from B to DA
            - The perpendicular distance from C to DA
            - The distance between B and C, parallel to DA.

        Args:
            a (float, float): The (column, row) position of point A.
            b (float, float): The (column, row) position of point B.
            c (float, float): The (column, row) position of point C.
            d (float, float): The (column, row) position of point D.
            **kwargs: See `new_roi` for all additional parameters.

        Returns:
             The created ROI instance.
        """
        points = np.vstack([np.array(d),
                            np.array(a),
                            np.array((0, 0)),  # Projection point of C onto AD calculated by OsiriX.
                            np.array(c),
                            np.array((0, 0)),  # Projection point of B onto AD calculated by OsiriX.
                            np.array(b)])
        return self.new_roi(points=points, itype=29, **kwargs)

    @pyosirix_connection_check
    def new_oval_angle_roi(self, center: Tuple[float, float], width: float,
                           height: float, **kwargs) -> osirix.roi.ROI:
        """ Create a new angle oval ROI within the viewer.

        Note: At present we are not aware that it is possible to pre-determine the angle part of
        this ROI. This method defines only the oval that will be drawn; the angle appears to have a
        default value of 113 degrees that can then be adapted by the user.

        Args:
            center (float, float): The (column, row) center of the oval ROI.
            width (float): The width of the oval ROI.
            height (float): The height of the oval ROI.
            **kwargs: See `new_roi` for all additional parameters.

        Returns:
             The created ROI instance.
        """
        rect = np.array([center[0], center[1], width / 2, height / 2])
        return self.new_roi(rect=rect, itype=31, **kwargs)
